@@ -163,13 +163,60 @@ static void cc_miller(element_t res, mpz_t q, point_t P,
     element_clear(e1);
 }
 
+static void cc_tatepower(element_ptr out, element_ptr in, pairing_t pairing)
+{
+    mnt_pairing_data_ptr p = pairing->data;
+    //TODO: this should only be called by odd k routines, so remove
+    //following section after even k case is fast enough
+    if (p->k == 6) {
+	element_t e0, e1, e2, e3;
+	element_init(e0, out->field);
+	element_init(e1, out->field);
+	element_init(e2, out->field);
+	element_init(e3, out->field);
+	void subexpr(element_ptr e) {
+	    element_set0(e0);
+	    poly_set_coeff(e0, poly_coeff(in, 0), 0);
+	    poly_const_mul(e2, poly_coeff(in, 1), e);
+	    element_add(e0, e0, e2);
+	    element_mul(e1, e, e);
+	    poly_const_mul(e2, poly_coeff(in, 2), e1);
+	    element_add(e0, e0, e2);
+	    element_mul(e1, e1, e);
+	    poly_const_mul(e2, poly_coeff(in, 3), e1);
+	    element_add(e0, e0, e2);
+	    element_mul(e1, e1, e);
+	    poly_const_mul(e2, poly_coeff(in, 4), e1);
+	    element_add(e0, e0, e2);
+	    element_mul(e1, e1, e);
+	    poly_const_mul(e2, poly_coeff(in, 5), e1);
+	    element_add(e0, e0, e2);
+	}
+	subexpr(p->xpowq[4]);
+	element_set(e3, e0);
+	subexpr(p->xpowq[3]);
+	element_mul(e3, e3, e0);
+	subexpr(p->xpowq[1]);
+	element_mul(e0, e0, in);
+	element_invert(e0, e0);
+	element_mul(out, e3, e0);
+	element_pow(out, out, p->tateexp);
+	element_clear(e0);
+	element_clear(e1);
+	element_clear(e2);
+	element_clear(e3);
+    } else {
+	element_pow(out, in, p->tateexp);
+    }
+}
+
 static void cc_pairing(element_ptr out, element_ptr in1, element_ptr in2,
 	pairing_t pairing)
 {
     mnt_pairing_data_ptr p = pairing->data;
     point_ptr Q = in2->data;
     cc_miller(out, pairing->r, in1->data, Q->x, Q->y, p->mapbase);
-    element_pow(out, out, p->tateexp);
+    cc_tatepower(out, out, pairing);
 }
 
 static int cc_is_almost_coddh_odd_k(element_ptr a, element_ptr b,
@@ -187,8 +234,8 @@ static int cc_is_almost_coddh_odd_k(element_ptr a, element_ptr b,
     point_ptr Pd = d->data;
     cc_miller(t0, pairing->r, a->data, Pd->x, Pd->y, p->mapbase);
     cc_miller(t1, pairing->r, b->data, Pc->x, Pc->y, p->mapbase);
-    element_pow(t0, t0, p->tateexp);
-    element_pow(t1, t1, p->tateexp);
+    cc_tatepower(t1, t1, pairing);
+    cc_tatepower(t0, t0, pairing);
     element_mul(t2, t0, t1);
     if (element_is1(t2)) {
 	//g, g^x, h, h^-x case
@@ -256,9 +303,33 @@ static void pairing_init_cc_param_odd_k(pairing_t pairing, cc_param_t param)
     field_init_polymod(p->Fqk, irred);
     element_clear(irred);
 
-    mpz_init(p->tateexp);
-    mpz_sub_ui(p->tateexp, p->Fqk->order, 1);
-    mpz_divexact(p->tateexp, p->tateexp, pairing->r);
+    if (param->k == 6) {
+	//TODO: remove this eventually, once even k case is fast enough
+	int i;
+	element_t e0;
+	mpz_ptr q = param->q;
+	mpz_ptr z = p->tateexp;
+	mpz_init(z);
+	mpz_mul(z, q, q);
+	mpz_sub(z, z, q);
+	mpz_add_ui(z, z, 1);
+	mpz_divexact(z, z, pairing->r);
+
+	p->xpowq = malloc(sizeof(element_ptr) * 5);
+	element_init(e0, p->Fqk);
+	poly_setx(e0);
+	for (i=1; i<=4; i++) {
+	    element_ptr e = p->xpowq[i] = malloc(sizeof(element_t));
+	    element_init(e, p->Fqk);
+	    element_pow(e0, e0, q);
+	    element_set(e, e0);
+	}
+	element_clear(e0);
+    } else {
+	mpz_init(p->tateexp);
+	mpz_sub_ui(p->tateexp, p->Fqk->order, 1);
+	mpz_divexact(p->tateexp, p->tateexp, pairing->r);
+    }
 
     p->mapbase = ((polymod_field_data_ptr) p->Fqk->data)->mapbase;
 
@@ -285,24 +356,24 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
 {
     //collate divisions
     int m;
-    element_t v, vd;
+    element_t v;
     point_t Z;
     element_t a, b, c;
     common_curve_ptr cc = P->curve->data;
     element_t t0;
     element_t e0, e1;
 
-    void do_vertical(element_t e)
+    void do_vertical()
     {
 	if (point_is_inf(Z)) {
 	    return;
 	}
 	mapbase(e0, Z->x);
 	element_sub(e0, Qx, e0);
-	element_mul(e, e, e0);
+	element_mul(v, v, e0);
     }
 
-    void do_tangent(element_t e)
+    void do_tangent()
     {
 	//a = -slope_tangent(Z.x, Z.y);
 	//b = 1;
@@ -323,7 +394,7 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
 
 	if (element_is0(Zy)) {
 	    //order 2
-	    do_vertical(e);
+	    do_vertical();
 	    return;
 	}
 
@@ -348,10 +419,10 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
 	element_add(e0, e0, e1);
 	mapbase(e1, c);
 	element_add(e0, e0, e1);
-	element_mul(e, e, e0);
+	element_mul(v, v, e0);
     }
 
-    void do_line(element_ptr e)
+    void do_line()
     {
 	//a = -(B.y - A.y) / (B.x - A.x);
 	//b = 1;
@@ -365,16 +436,16 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
 
 	//we assume B is never O
 	if (point_is_inf(Z)) {
-	    do_vertical(e);
+	    do_vertical();
 	    return;
 	}
 
 	if (!element_cmp(Ax, Bx)) {
 	    if (!element_cmp(Ay, By)) {
-		do_tangent(e);
+		do_tangent();
 		return;
 	    }
-	    do_vertical(e);
+	    do_vertical();
 	    return;
 	}
 
@@ -392,7 +463,7 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
 	element_add(e0, e0, e1);
 	mapbase(e1, c);
 	element_add(e0, e0, e1);
-	element_mul(e, e, e0);
+	element_mul(v, v, e0);
     }
 
     element_init(a, P->curve->field);
@@ -403,21 +474,19 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     element_init(e1, res->field);
 
     element_init(v, res->field);
-    element_init(vd, res->field);
     point_init(Z, P->curve);
 
     point_set(Z, P);
 
     element_set1(v);
-    element_set1(vd);
     m = mpz_sizeinbase(q, 2) - 2;
 
     while(m >= 0) {
 	element_mul(v, v, v);
-	do_tangent(v);
+	do_tangent();
 	point_double(Z, Z);
 	if (mpz_tstbit(q, m)) {
-	    do_line(v);
+	    do_line();
 	    point_add(Z, Z, P);
 	}
 	m--;
@@ -426,7 +495,6 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     element_set(res, v);
 
     element_clear(v);
-    element_clear(vd);
     point_clear(Z);
     element_clear(a);
     element_clear(b);
@@ -436,6 +504,53 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     element_clear(e1);
 }
 
+static void cc_tatepower_even_k(element_ptr out, element_ptr in, pairing_t pairing)
+{
+    even_mnt_pairing_data_ptr p = pairing->data;
+    if (p->k == 6) {
+	element_t e0, e1, e2, e3;
+	element_init(e0, p->Fqk);
+	element_init(e1, p->Fqd);
+	element_init(e2, p->Fqd);
+	element_init(e3, p->Fqk);
+	element_ptr e0re = fi_re(e0);
+	element_ptr e0im = fi_im(e0);
+	element_ptr inre = fi_re(in);
+	element_ptr inim = fi_im(in);
+	void subexpr(element_ptr e) {
+	    element_set0(e0);
+	    poly_set_coeff(e0re, poly_coeff(inre, 0), 0);
+	    poly_set_coeff(e0im, poly_coeff(inim, 0), 0);
+	    poly_const_mul(e2, poly_coeff(inre, 1), e);
+	    element_add(e0re, e0re, e2);
+	    poly_const_mul(e2, poly_coeff(inim, 1), e);
+	    element_add(e0im, e0im, e2);
+	    element_mul(e1, e, e);
+	    poly_const_mul(e2, poly_coeff(inre, 2), e1);
+	    element_add(e0re, e0re, e2);
+	    poly_const_mul(e2, poly_coeff(inim, 2), e1);
+	    element_add(e0im, e0im, e2);
+	}
+	subexpr(p->xpowq[4]);
+	element_set(e3, e0);
+	subexpr(p->xpowq[3]);
+	element_neg(e0im, e0im);
+	element_mul(e3, e3, e0);
+	subexpr(p->xpowq[1]);
+	element_neg(e0im, e0im);
+	element_mul(e0, e0, in);
+	element_invert(e0, e0);
+	element_mul(out, e3, e0);
+	element_pow(out, out, p->tateexp);
+	element_clear(e0);
+	element_clear(e1);
+	element_clear(e2);
+	element_clear(e3);
+    } else {
+	element_pow(out, in, p->tateexp);
+    }
+}
+
 static void cc_pairing_even_k(element_ptr out, element_ptr in1, element_ptr in2,
 	pairing_t pairing)
 {
@@ -451,7 +566,7 @@ static void cc_pairing_even_k(element_ptr out, element_ptr in1, element_ptr in2,
     //v^-3/2 = v^-2 * v^1/2
     element_mul(fi_im(y), Qbase->y, p->nqrinv2);
     cc_miller_no_denom(out, pairing->r, in1->data, x, y, p->mapbase);
-    element_pow(out, out, p->tateexp);
+    cc_tatepower_even_k(out, out, pairing);
 }
 
 static int cc_is_almost_coddh_even_k(element_ptr a, element_ptr b,
@@ -483,9 +598,9 @@ static int cc_is_almost_coddh_even_k(element_ptr a, element_ptr b,
     element_mul(fi_im(dy), Pd->y, p->nqrinv2);
 
     cc_miller_no_denom(t0, pairing->r, a->data, dx, dy, p->mapbase);
-    cc_miller(t1, pairing->r, b->data, cx, cy, p->mapbase);
-    element_pow(t0, t0, p->tateexp);
-    element_pow(t1, t1, p->tateexp);
+    cc_miller_no_denom(t1, pairing->r, b->data, cx, cy, p->mapbase);
+    cc_tatepower_even_k(t0, t0, pairing);
+    cc_tatepower_even_k(t1, t1, pairing);
     element_mul(t2, t0, t1);
     if (element_is1(t2)) {
 	//g, g^x, h, h^-x case
@@ -506,12 +621,8 @@ static int cc_is_almost_coddh_even_k(element_ptr a, element_ptr b,
 
 static void Fq_to_Fqd_to_Fqk(element_t out, element_t in)
 {
-    element_t e0;
-    fq_field_data_ptr p = out->field->data;
-    element_init(e0, p->field);
-    element_field_to_poly(e0, in);
-    element_field_to_quadratic(out, e0);
-    element_clear(e0);
+    element_field_to_poly(fi_re(out), in);
+    element_set0(fi_im(out));
 }
 
 static void pairing_init_cc_param_even_k(pairing_t pairing, cc_param_t param)
@@ -543,11 +654,45 @@ static void pairing_init_cc_param_even_k(pairing_t pairing, cc_param_t param)
     } while (!poly_is_irred(irred));
     field_init_polymod(p->Fqd, irred);
     element_clear(irred);
+
+    //find a quadratic nonresidue of Fqd lying in Fq
+    //TODO: assuming exists an element in Fq that is not a square in Fqd
+    //assert(!p->Fqd->nqr);
+    p->Fqd->nqr = malloc(sizeof(element_t));
+    element_init(p->Fqd->nqr, p->Fqd);
+    do {
+	poly_alloc(p->Fqd->nqr, 1);
+	element_random(poly_coeff(p->Fqd->nqr, 0));
+    } while (element_is_sqr(p->Fqd->nqr));
+
     field_init_quadratic(p->Fqk, p->Fqd);
 
-    mpz_init(p->tateexp);
-    mpz_sub_ui(p->tateexp, p->Fqk->order, 1);
-    mpz_divexact(p->tateexp, p->tateexp, pairing->r);
+    if (param->k == 6) {
+	int i;
+	element_t e0;
+	mpz_ptr q = param->q;
+	mpz_ptr z = p->tateexp;
+	mpz_init(z);
+	mpz_mul(z, q, q);
+	mpz_sub(z, z, q);
+	mpz_add_ui(z, z, 1);
+	mpz_divexact(z, z, pairing->r);
+
+	p->xpowq = malloc(sizeof(element_ptr) * 5);
+	element_init(e0, p->Fqd);
+	poly_setx(e0);
+	for (i=1; i<=4; i++) {
+	    element_ptr e = p->xpowq[i] = malloc(sizeof(element_t));
+	    element_init(e, p->Fqd);
+	    element_pow(e0, e0, q);
+	    element_set(e, e0);
+	}
+	element_clear(e0);
+    } else {
+	mpz_init(p->tateexp);
+	mpz_sub_ui(p->tateexp, p->Fqk->order, 1);
+	mpz_divexact(p->tateexp, p->tateexp, pairing->r);
+    }
 
     p->mapbase = Fq_to_Fqd_to_Fqk;
 
@@ -577,8 +722,12 @@ static void pairing_init_cc_param_even_k(pairing_t pairing, cc_param_t param)
 
 void pairing_init_cc_param(pairing_t pairing, cc_param_t param)
 {
-    if (param->k % 2) pairing_init_cc_param_odd_k(pairing, param);
-    else pairing_init_cc_param_even_k(pairing, param);
+    if (0) {
+	pairing_init_cc_param_odd_k(pairing, param);
+    } else {
+	if (param->k % 2) pairing_init_cc_param_odd_k(pairing, param);
+	else pairing_init_cc_param_even_k(pairing, param);
+    }
 }
 
 static void phi_identity(element_ptr out, element_ptr in, pairing_ptr pairing)
@@ -660,6 +809,7 @@ static void solinas_pairing(element_ptr out, element_ptr in1, element_ptr in2,
 	//check against standard Miller's alogrithm
 	fq_data_ptr fdp;
 	element_t x, y;
+	mpz_t exp;
 
 	element_init(x, p->Fq2);
 	element_init(y, p->Fq2);
@@ -671,7 +821,12 @@ static void solinas_pairing(element_ptr out, element_ptr in1, element_ptr in2,
 	element_set0(fdp->x);
 
 	cc_miller(out, pairing->r, in1->data, x, y, element_field_to_quadratic);
-	element_pow(out, out, p->tateexp);
+
+	mpz_init(exp);
+	mpz_sub_ui(exp, p->Fq2->order, 1);
+	mpz_divexact(exp, exp, pairing->r);
+	element_pow(out, out, exp);
+	mpz_clear(exp);
 	printf("check: ");
 	element_out_str(stdout, 0, out);
 	printf("\n");
@@ -763,10 +918,6 @@ void pairing_init_solinas_param(pairing_t pairing, solinas_param_t param)
 
     field_init_fi(p->Fq2, p->Fq);
 
-    mpz_init(p->tateexp);
-    mpz_sub_ui(p->tateexp, p->Fq2->order, 1);
-    mpz_divexact(p->tateexp, p->tateexp, pairing->r);
-
     mpz_init(p->h);
     mpz_set(p->h, param->h);
 
