@@ -30,6 +30,7 @@ struct mnt_pairing_data_s {
     mpz_t tateexp;
     int k;
     fieldmap mapbase;
+    element_ptr *xpowq;
 };
 typedef struct mnt_pairing_data_s mnt_pairing_data_t[1];
 typedef struct mnt_pairing_data_s *mnt_pairing_data_ptr;
@@ -41,6 +42,7 @@ struct even_mnt_pairing_data_s {
     mpz_t tateexp;
     int k;
     fieldmap mapbase;
+    element_ptr *xpowq;
 };
 typedef struct even_mnt_pairing_data_s even_mnt_pairing_data_t[1];
 typedef struct even_mnt_pairing_data_s *even_mnt_pairing_data_ptr;
@@ -48,7 +50,6 @@ typedef struct even_mnt_pairing_data_s *even_mnt_pairing_data_ptr;
 struct solinas_pairing_data_s {
     field_t Fq, Fq2;
     curve_t Eq, Eq2;
-    mpz_t tateexp; //TODO: get rid of?
     mpz_t h;
     int exp2, exp1;
     int sign1;
