@@ -436,7 +436,7 @@ static void polymod_invert(element_ptr r, element_ptr f)
     poly_invert(r, f, p->poly);
 }
 
-static void poly_basemap(element_ptr f, element_ptr g)
+void element_field_to_poly(element_ptr f, element_ptr g)
 {
     poly_alloc(f, 1);
     element_set(poly_coeff(f, 0), g);
@@ -469,7 +469,7 @@ void field_init_poly(field_ptr f, field_ptr base_field)
     f->data = malloc(sizeof(poly_field_data_t));
     p = f->data;
     p->field = base_field;
-    p->mapbase = poly_basemap;
+    p->mapbase = element_field_to_poly;
     f->field_clear = field_clear_poly;
     f->init = poly_init;
     f->clear = poly_clear;
@@ -707,7 +707,7 @@ void field_init_polymod(field_ptr f, element_ptr poly)
     f->data = malloc(sizeof(polymod_field_data_t));
     p = f->data;
     p->field = pdp->field;
-    p->mapbase = poly_basemap;
+    p->mapbase = element_field_to_poly;
     element_init(p->poly, poly->field);
     element_set(p->poly, poly);
     f->field_clear = field_clear_polymod;
@@ -733,11 +733,11 @@ void field_init_polymod(field_ptr f, element_ptr poly)
     f->sqrt = polymod_sqrt;
     f->to_bytes = polymod_to_bytes;
     f->from_bytes = polymod_from_bytes;
-    if (poly->field->fixed_length_in_bytes < 0) {
+    if (pdp->field->fixed_length_in_bytes < 0) {
 	f->fixed_length_in_bytes = -1;
 	f->length_in_bytes = polymod_length_in_bytes;
     } else {
-	f->fixed_length_in_bytes = p->field->fixed_length_in_bytes * poly_degree(poly);
+	f->fixed_length_in_bytes = pdp->field->fixed_length_in_bytes * poly_degree(poly);
     }
     n = pd->coeff->count;
     mpz_pow_ui(f->order, p->field->order, n - 1);
